@@ -13,12 +13,16 @@
 #include "../src/syscall.h"
 #include "nvme.h"
 
+#define min(a, b)	((a) < (b) ? (a) : (b))
+
 #define FILE_SIZE	(256 * 1024)
 #define BS		8192
 #define BUFFERS		(FILE_SIZE / BS)
 
-static struct iovec *vecs;
+static void *meta_mem;
+static struct iovec *vecs, *backing_vec;
 static int no_pt;
+static bool vec_fixed_supported = true;
 
 /*
  * Each offset in the file has the ((test_case / 2) * FILE_SIZE)
@@ -66,14 +70,14 @@ static int fill_pattern(int tc)
 }
 
 static int __test_io(const char *file, struct io_uring *ring, int tc, int read,
-		     int sqthread, int fixed, int nonvec)
+		     int sqthread, int fixed, int nonvec, int async, int linked)
 {
 	struct io_uring_sqe *sqe;
 	struct io_uring_cqe *cqe;
 	struct nvme_uring_cmd *cmd;
 	int open_flags;
 	int do_fixed;
-	int i, ret, fd = -1;
+	int i, ret, fd = -1, use_fd = -1, submit_count = 0;
 	off_t offset;
 	__u64 slba;
 	__u32 nlb;
@@ -84,7 +88,7 @@ static int __test_io(const char *file, struct io_uring *ring, int tc, int read,
 		open_flags = O_WRONLY;
 
 	if (fixed) {
-		ret = t_register_buffers(ring, vecs, BUFFERS);
+		ret = t_register_buffers(ring, backing_vec, 1);
 		if (ret == T_SETUP_SKIP)
 			return 0;
 		if (ret != T_SETUP_OK) {
@@ -114,59 +118,39 @@ static int __test_io(const char *file, struct io_uring *ring, int tc, int read,
 
 	offset = 0;
 	for (i = 0; i < BUFFERS; i++) {
+		unsigned int iovcnt = 1;
+		size_t total_len;
+
+		if (linked) {
+			sqe = io_uring_get_sqe(ring);
+			io_uring_prep_nop(sqe);
+			if (async)
+				sqe->flags |= IOSQE_ASYNC;
+			sqe->flags |= IOSQE_IO_LINK;
+			sqe->user_data = 0x1000;
+			submit_count++;
+		}
+
 		sqe = io_uring_get_sqe(ring);
 		if (!sqe) {
 			fprintf(stderr, "sqe get failed\n");
 			goto err;
 		}
-		if (read) {
-			int use_fd = fd;
+		use_fd = fd;
+		do_fixed = fixed;
 
-			do_fixed = fixed;
-
-			if (sqthread)
-				use_fd = 0;
-			if (fixed && (i & 1))
-				do_fixed = 0;
-			if (do_fixed) {
-				io_uring_prep_read_fixed(sqe, use_fd, vecs[i].iov_base,
-								vecs[i].iov_len,
-								offset, i);
-				sqe->cmd_op = NVME_URING_CMD_IO;
-			} else if (nonvec) {
-				io_uring_prep_read(sqe, use_fd, vecs[i].iov_base,
-							vecs[i].iov_len, offset);
-				sqe->cmd_op = NVME_URING_CMD_IO;
-			} else {
-				io_uring_prep_readv(sqe, use_fd, &vecs[i], 1,
-								offset);
-				sqe->cmd_op = NVME_URING_CMD_IO_VEC;
-			}
-		} else {
-			int use_fd = fd;
-
-			do_fixed = fixed;
-
-			if (sqthread)
-				use_fd = 0;
-			if (fixed && (i & 1))
-				do_fixed = 0;
-			if (do_fixed) {
-				io_uring_prep_write_fixed(sqe, use_fd, vecs[i].iov_base,
-								vecs[i].iov_len,
-								offset, i);
-				sqe->cmd_op = NVME_URING_CMD_IO;
-			} else if (nonvec) {
-				io_uring_prep_write(sqe, use_fd, vecs[i].iov_base,
-							vecs[i].iov_len, offset);
-				sqe->cmd_op = NVME_URING_CMD_IO;
-			} else {
-				io_uring_prep_writev(sqe, use_fd, &vecs[i], 1,
-								offset);
-				sqe->cmd_op = NVME_URING_CMD_IO_VEC;
-			}
-		}
-		sqe->opcode = IORING_OP_URING_CMD;
+		if (sqthread)
+			use_fd = 0;
+		if (nonvec)
+			io_uring_prep_uring_cmd(sqe, NVME_URING_CMD_IO, use_fd);
+		else
+			io_uring_prep_uring_cmd(sqe, NVME_URING_CMD_IO_VEC, use_fd);
+		if (fixed && (i & 1))
+			do_fixed = 0;
+		if (do_fixed)
+			sqe->buf_index = 0;
+		if (async)
+			sqe->flags |= IOSQE_ASYNC;
 		if (do_fixed)
 			sqe->uring_cmd_flags |= IORING_URING_CMD_FIXED;
 		sqe->user_data = ((uint64_t)offset << 32) | i;
@@ -178,39 +162,60 @@ static int __test_io(const char *file, struct io_uring *ring, int tc, int read,
 
 		cmd->opcode = read ? nvme_cmd_read : nvme_cmd_write;
 
+		if (!nonvec) {
+			iovcnt = (submit_count % 3 == 0) ? 1 : ((submit_count % 3 == 1) ? 3 : 9);
+			iovcnt = min(iovcnt, BUFFERS - i);
+		}
+		total_len = BS * iovcnt;
+
 		slba = offset >> lba_shift;
-		nlb = (BS >> lba_shift) - 1;
+		nlb = (total_len >> lba_shift) - 1;
 
 		/* cdw10 and cdw11 represent starting lba */
 		cmd->cdw10 = slba & 0xffffffff;
 		cmd->cdw11 = slba >> 32;
 		/* cdw12 represent number of lba's for read/write */
 		cmd->cdw12 = nlb;
-		if (do_fixed || nonvec) {
+		if (nonvec) {
 			cmd->addr = (__u64)(uintptr_t)vecs[i].iov_base;
 			cmd->data_len = vecs[i].iov_len;
 		} else {
 			cmd->addr = (__u64)(uintptr_t)&vecs[i];
-			cmd->data_len = 1;
+			cmd->data_len = iovcnt;
+		}
+
+		if (meta_size) {
+			cmd->metadata = (__u64)(uintptr_t)(meta_mem +
+						meta_size * i * (nlb + 1));
+			cmd->metadata_len = meta_size * (nlb + 1);
 		}
 		cmd->nsid = nsid;
 
-		offset += BS;
+		offset += total_len;
+		if (!nonvec)
+			i += iovcnt - 1;
+		submit_count++;
 	}
 
 	ret = io_uring_submit(ring);
-	if (ret != BUFFERS) {
+	if (ret != submit_count) {
 		fprintf(stderr, "submit got %d, wanted %d\n", ret, BUFFERS);
 		goto err;
 	}
 
-	for (i = 0; i < BUFFERS; i++) {
+	for (i = 0; i < submit_count; i++) {
+		int is_link;
+
 		ret = io_uring_wait_cqe(ring, &cqe);
 		if (ret) {
 			fprintf(stderr, "wait_cqe=%d\n", ret);
 			goto err;
 		}
 		if (cqe->res != 0) {
+			if (cqe->res == -EINVAL && fixed && !nonvec) {
+				vec_fixed_supported = false;
+				goto cleanup_and_skip;
+			}
 			if (!no_pt) {
 				no_pt = 1;
 				goto skip;
@@ -218,7 +223,10 @@ static int __test_io(const char *file, struct io_uring *ring, int tc, int read,
 			fprintf(stderr, "cqe res %d, wanted 0\n", cqe->res);
 			goto err;
 		}
+		is_link = cqe->user_data == 0x1000;
 		io_uring_cqe_seen(ring, cqe);
+		if (is_link)
+			continue;
 		if (read) {
 			int index = cqe->user_data & 0xffffffff;
 			void *buf = vecs[index].iov_base;
@@ -229,6 +237,7 @@ static int __test_io(const char *file, struct io_uring *ring, int tc, int read,
 		}
 	}
 
+cleanup_and_skip:
 	if (fixed) {
 		ret = io_uring_unregister_buffers(ring);
 		if (ret) {
@@ -254,7 +263,7 @@ err:
 }
 
 static int test_io(const char *file, int tc, int read, int sqthread,
-		   int fixed, int nonvec, int hybrid)
+		   int fixed, int nonvec, int hybrid, int async, int linked)
 {
 	struct io_uring ring;
 	int ret, ring_flags = 0;
@@ -268,7 +277,10 @@ static int test_io(const char *file, int tc, int read, int sqthread,
 	if (hybrid)
 		ring_flags |= IORING_SETUP_IOPOLL | IORING_SETUP_HYBRID_IOPOLL;
 
-	ret = t_create_ring(64, &ring, ring_flags);
+	if (fixed && (!vec_fixed_supported && !nonvec))
+		return 0;
+
+	ret = t_create_ring(128, &ring, ring_flags);
 	if (ret == T_SETUP_SKIP)
 		return 0;
 	if (ret != T_SETUP_OK) {
@@ -280,7 +292,7 @@ static int test_io(const char *file, int tc, int read, int sqthread,
 		return 1;
 	}
 
-	ret = __test_io(file, &ring, tc, read, sqthread, fixed, nonvec);
+	ret = __test_io(file, &ring, tc, read, sqthread, fixed, nonvec, async, linked);
 	io_uring_queue_exit(&ring);
 
 	return ret;
@@ -314,9 +326,7 @@ static int test_invalid_passthru_submit(const char *file)
 	}
 
 	sqe = io_uring_get_sqe(&ring);
-	io_uring_prep_read(sqe, fd, vecs[0].iov_base, vecs[0].iov_len, 0);
-	sqe->cmd_op = NVME_URING_CMD_IO;
-	sqe->opcode = IORING_OP_URING_CMD;
+	io_uring_prep_uring_cmd(sqe, NVME_URING_CMD_IO, fd);
 	sqe->user_data = 1;
 	cmd = (struct nvme_uring_cmd *)sqe->cmd;
 	memset(cmd, 0, sizeof(struct nvme_uring_cmd));
@@ -387,10 +397,8 @@ static int test_io_uring_submit_enters(const char *file)
 		__u32 nlb;
 
 		sqe = io_uring_get_sqe(&ring);
-		io_uring_prep_readv(sqe, fd, &vecs[i], 1, offset);
+		io_uring_prep_uring_cmd(sqe, NVME_URING_CMD_IO, fd);
 		sqe->user_data = i;
-		sqe->opcode = IORING_OP_URING_CMD;
-		sqe->cmd_op = NVME_URING_CMD_IO;
 		cmd = (struct nvme_uring_cmd *)sqe->cmd;
 		memset(cmd, 0, sizeof(struct nvme_uring_cmd));
 
@@ -404,6 +412,12 @@ static int test_io_uring_submit_enters(const char *file)
 		cmd->addr = (__u64)(uintptr_t)&vecs[i];
 		cmd->data_len = 1;
 		cmd->nsid = nsid;
+
+		if (meta_size) {
+			cmd->metadata = (__u64)(uintptr_t)(meta_mem +
+						meta_size * i * (nlb + 1));
+			cmd->metadata_len = meta_size * (nlb + 1);
+		}
 	}
 
 	/* submit manually to avoid adding IORING_ENTER_GETEVENTS */
@@ -450,16 +464,27 @@ int main(int argc, char *argv[])
 	if (ret)
 		return T_EXIT_SKIP;
 
-	vecs = t_create_buffers(BUFFERS, BS);
+	vecs = t_malloc(BUFFERS * sizeof(struct iovec));
+	backing_vec = t_create_buffers(1, BUFFERS * BS);
+	/* Slice single large backing_vec into multiple smaller vecs */
+	for (int i = 0; i < BUFFERS; i++) {
+		vecs[i].iov_base = backing_vec[0].iov_base + i * BS;
+		vecs[i].iov_len = BS;
+	}
+	if (meta_size)
+		t_posix_memalign(&meta_mem, 0x1000,
+				 meta_size * BUFFERS * (BS >> lba_shift));
 
-	for (i = 0; i < 32; i++) {
+	for (i = 0; i < 64; i++) {
 		int read = (i & 1) != 0;
 		int sqthread = (i & 2) != 0;
 		int fixed = (i & 4) != 0;
 		int nonvec = (i & 8) != 0;
 		int hybrid = (i & 16) != 0;
+		int async = (i & 32) != 0;
+		int linked = (i & 64) != 0;
 
-		ret = test_io(fname, i, read, sqthread, fixed, nonvec, hybrid);
+		ret = test_io(fname, i, read, sqthread, fixed, nonvec, hybrid, async, linked);
 		if (no_pt)
 			break;
 		if (ret) {
